#ifndef CLUSTER_H
#define CLUSTER_H

#include "tree.h"
#include "iterator.h"
#include <vector>
#include <list>
#include <map>
#include <iostream>
#include <algorithm>

// Namespace Cluster, with all
namespace cluster {


//! assuming that there is less than 1<<16 groups ?
typedef unsigned GroupID;

class Observation;

/**
	This nested class allow to make
	the Composite design pattern implementation
	
	Thus, A group or an observation should be at the
	same level (in fact, use the same interface)
*/
class Nested
{
  public:
	Nested() {}
	virtual ~Nested(){}

	// Information
	virtual void      print(std::ostream& str = std::cout) const = 0;
	virtual unsigned  size()    = 0;
	virtual double    compact() = 0;
};


/**
	A Group is a composition of several observations
	Theses observations may be Group or real observations
	(Evolutive structure)
*/
class Group : public Nested
{
	typedef std::list<Nested *>::iterator iterator;

	std::list<Nested *> observations;
  public:

	Group() {};
	Group(const Group& g);

	// Some function to add observation in Group
	void add   (Nested *);
	void erase (Nested *);
	
	inline void add(const Observation& o) {
		Observation* ptr = &(const_cast<Observation&>(o));
		add(reinterpret_cast<Nested *>(ptr));
	}
	
	inline iterator begin() { return observations.begin(); }
	inline iterator end()   { return observations.end();   }

	// Nested's specialization
  public:

	void     print(std::ostream& str = std::cout) const;
	unsigned size();
	double   compact();

	friend std::ostream& operator<<(std::ostream& str, const Group& n);
};

/**
	An observation is a token that compose groups/clusters
*/
class Observation : public Nested
{
  public:
	unsigned id; //!< The number of the current observation

	Observation(unsigned _id = 0) : id(_id) {}

	// Nested's specialization
  public:

	inline void     print(std::ostream& str = std::cout) const { str << id; }
	inline unsigned size()    { return 1  ; }
	inline double   compact() { return 1.0; }

	friend std::ostream& operator<<(std::ostream& str, const Observation& n);
};


// Useful function calling specialized print
std::ostream& operator<<(std::ostream& str, const Group& n);
std::ostream& operator<<(std::ostream& str, const Observation& n);


/**
	This class contains a full cluster
	A cluster is an union of Groups which
	contain observations
*/
class Cluster
{
  public:
    typedef GroupIterator  iterator;

  private:
	tree<GroupID>         cluster; //!< A tree base on the ID of groups
	std::vector<Group *>  groups;  //!< Vector on pointers to real groups (no deprecated)

	// Get reversed information about ID and Group *
	std::map<Group *, GroupID>  ptr2ID;
	std::map<GroupID,  Group *> id2Ptr;

  public:
	Cluster() {}
	~Cluster();

	inline tree<GroupID>& getCluster() { return cluster; }

	//! Add a group to a cluster
	void add(Group *g);
	void add(GroupIterator iter);

	//! Return the number of groups in the cluster
	inline unsigned size() {
		return groups.size();
	}

	//! Add an observation to a cluster <=> add a deprecated group 
	//! with only one observation
	inline void add(Observation *o) {
		Group g;
		g.add(o);
		add(new Group(g));
	}

	//! Delete a group if it exist
	void remove(Group *g);
	void remove(GroupIterator iter);

	//! Test if a cluster is the same as this
	inline bool operator==(const Cluster& c) const {
		if (c.groups.begin() == groups.begin() 
		&&  c.groups.end()   == groups.end())
			return true;

		return false;
	}

	//! Difference test (Cluster is not Cluster)
	inline bool operator!=(const Cluster& c) const {return !(this->operator==(c));}

	// iterator information
	inline iterator begin() { return iterator(groups.begin()); }
	inline iterator end()   { return iterator(groups.end());   }


	//! Print the current Cluster
	void print(std::ostream& str = std::cout) const;

	friend std::ostream& operator<<(std::ostream& str, const Cluster& c);

};

std::ostream& operator<<(std::ostream& str, const Cluster& c);


}

#endif
